test = "TEST"
__version__ = '0.0.10'
# try:
#     import pkg_resources
#     __version__ = pkg_resources.safe_version(pkg_resources.get_distribution(__name__).version)
# except:
#     __version__ = 'unknown'
